@@ -2,12 +2,12 @@
  Adapted from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/Terminal.py
 """
 
-from pyqtgraph.Qt import QtCore, QtGui
+from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
 from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject
 from pyqtgraph import functions as fn
 from pyqtgraph.Point import Point
 
-from eagerx.core import constants
+from eagerx_gui import configuration
 from eagerx_gui.pyqtgraph_utils import (
     exception_handler,
     ConnectionDialog,
@@ -20,8 +20,6 @@ from eagerx.utils.utils import (
     get_attribute_from_module,
 )
 from eagerx.core.converters import Identity
-from eagerx.core.specs import EntitySpec
-from eagerx.core.view import SpecView
 
 
 class GuiTerminal(object):
@@ -37,12 +35,12 @@ class GuiTerminal(object):
         self.terminal_type = name_split[0]
         self.terminal_name = name_split[-1]
 
-        assert self.terminal_type in set.union(constants.TERMS_IN, constants.TERMS_OUT), (
+        assert self.terminal_type in set.union(configuration.TERMS_IN, configuration.TERMS_OUT), (
             f"Invalid terminal type: {self.terminal_type}, "
-            f"should be one of {set.union(constants.TERMS_IN, constants.TERMS_OUT)}"
+            f"should be one of {set.union(configuration.TERMS_IN, configuration.TERMS_OUT)}"
         )
 
-        self.is_input = self.terminal_type in constants.TERMS_IN
+        self.is_input = self.terminal_type in configuration.TERMS_IN
         self.is_state = self.terminal_type in ["targets", "states"]
         self.is_feedthrough = self.terminal_type == "feedthroughs"
         self.is_removable = node.is_object or self.node_type in [
@@ -50,7 +48,7 @@ class GuiTerminal(object):
             "observations",
         ]
         self.is_renamable = self.node_type in ["actions", "observations"]
-        self.is_connectable = not (self.terminal_type == "states" and not node.is_object)
+        self.is_connectable = not (self.node_type == "engine_node" or (self.terminal_type == "states" and not node.is_object))
 
         self._graphicsItem = TerminalGraphicsItem(self, parent=self.node.graphics_item())
         self.recolor()
@@ -163,7 +161,7 @@ class GuiTerminal(object):
                     msg_type_in = get_opposite_msg_cls(target_params["msg_type"], target_params["converter"])
                 else:
                     msg_type_in = get_attribute_from_module(target_params["msg_type"])
-                space_converter = (target_params["space_converter"] if target_params["space_converter"] else identity_converter)
+                space_converter = target_params["space_converter"] if target_params["space_converter"] else identity_converter
                 space_converter_dialog = ConverterDialog(
                     space_converter,
                     parent,
@@ -224,6 +222,8 @@ class GuiTerminal(object):
                     color = QtGui.QColor(255, 0, 0)
                 else:
                     color = QtGui.QColor(128, 128, 128)
+            elif self.is_feedthrough:
+                color = QtGui.QColor(175, 238, 238)
             else:
                 color = QtGui.QColor(0, 0, 255)
         self.graphics_item().setBrush(QtGui.QBrush(color))
@@ -241,9 +241,9 @@ class GuiTerminal(object):
             self.graphics_item().term_renamed(self.name)
             return
         node_name, _, old_name = self.connection_tuple()
-        if node_name == 'env/actions':
+        if node_name == "env/actions":
             self.node.graph.rename(new=name.split("/")[-1], action=old_name)
-        elif node_name == 'env/observations':
+        elif node_name == "env/observations":
             self.node.graph.rename(new=name.split("/")[-1], observation=old_name)
         else:
             raise NotImplementedError("Can only change terminal names for actions and observations.")
@@ -266,20 +266,22 @@ class GuiTerminal(object):
             item.scene().removeItem(item)
 
 
-class TextItem(QtGui.QGraphicsTextItem):
-    def __init__(self, is_renamable, text, termimnal_graphics_item):
-        self._is_renamable = is_renamable
-        super().__init__(text, termimnal_graphics_item)
+class TextItem(QtWidgets.QGraphicsTextItem):
+    def __init__(self, text, parent, on_update):
+        super().__init__(text, parent)
+        self.on_update = on_update
 
-    def mousePressEvent(self, ev):
-        if ev.button() == QtCore.Qt.RightButton:
-            if self._is_renamable:
-                self.setTextInteractionFlags(QtCore.Qt.NoTextInteraction)
-            ev.ignore()
-        elif ev.button() == QtCore.Qt.LeftButton:
-            if self._is_renamable:
-                self.setTextInteractionFlags(QtCore.Qt.TextEditorInteraction)
-            super().mousePressEvent(ev)
+    def focusOutEvent(self, ev):
+        super().focusOutEvent(ev)
+        if self.on_update is not None:
+            self.on_update()
+
+    def keyPressEvent(self, ev):
+        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
+            if self.on_update is not None:
+                self.on_update()
+                return
+        super().keyPressEvent(ev)
 
 
 class TerminalGraphicsItem(GraphicsObject):
@@ -287,27 +289,28 @@ class TerminalGraphicsItem(GraphicsObject):
         self.term = term
         GraphicsObject.__init__(self, parent)
         self.brush = fn.mkBrush(0, 0, 0)
-        self.box = QtGui.QGraphicsRectItem(0, 0, 10, 10, self)
-        self.label = TextItem(self.term.is_renamable, self.term.terminal_name, self)
-        self.label.scale(0.7, 0.7)
+        self.box = QtWidgets.QGraphicsRectItem(0, 0, 10, 10, self)
+        on_update = self.label_changed if self.term.is_renamable else None
+        self.label = TextItem(self.term.terminal_name, self, on_update)
+        self.label.setScale(0.7)
         self.newConnection = None
         self.setFiltersChildEvents(True)  # to pick up mouse events on the rectitem
         if self.term.is_renamable:
-            self.label.setTextInteractionFlags(QtCore.Qt.TextEditorInteraction)
+            self.label.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextEditorInteraction)
             self.label.focusOutEvent = self.label_focus_out
             self.label.keyPressEvent = self.label_key_press
         self.setZValue(1)
         self.menu = None
 
     def label_focus_out(self, ev):
-        QtGui.QGraphicsTextItem.focusOutEvent(self.label, ev)
+        QtWidgets.QGraphicsTextItem.focusOutEvent(self.label, ev)
         self.label_changed()
 
     def label_key_press(self, ev):
-        if ev.key() == QtCore.Qt.Key_Enter or ev.key() == QtCore.Qt.Key_Return:
+        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
             self.label_changed()
         else:
-            QtGui.QGraphicsTextItem.keyPressEvent(self.label, ev)
+            QtWidgets.QGraphicsTextItem.keyPressEvent(self.label, ev)
 
     def label_changed(self):
         new_name = str(self.label.toPlainText())
@@ -374,15 +377,15 @@ class TerminalGraphicsItem(GraphicsObject):
         ev.ignore()  # necessary to allow click/drag events to process correctly
 
     def mouseClickEvent(self, ev):
-        if ev.button() == QtCore.Qt.LeftButton:
+        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
             ev.accept()
-            self.label.setFocus(QtCore.Qt.MouseFocusReason)
-        elif ev.button() == QtCore.Qt.RightButton:
+            self.label.setFocus(QtCore.Qt.FocusReason.MouseFocusReason)
+        elif ev.button() == QtCore.Qt.MouseButton.RightButton:
             ev.accept()
             self.raise_context_menu(ev)
 
     def mouseDoubleClickEvent(self, ev):
-        if int(ev.button()) == int(QtCore.Qt.LeftButton):
+        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
             ev.accept()
             param_window = ParamWindow(node=self.term.node, term=self.term)
             param_window.open()
@@ -395,7 +398,7 @@ class TerminalGraphicsItem(GraphicsObject):
         menu.popup(QtCore.QPoint(pos.x(), pos.y()))
 
     def get_menu(self):
-        self.menu = QtGui.QMenu()
+        self.menu = QtWidgets.QMenu()
         self.menu.setTitle("Terminal")
         rem_act = QtGui.QAction("Remove {}".format(self.term.terminal_name), self.menu)
         rem_act.triggered.connect(self.remove_self)
@@ -415,7 +418,7 @@ class TerminalGraphicsItem(GraphicsObject):
         self.term.node.remove_terminal(self.term)
 
     def mouseDragEvent(self, ev):
-        if ev.button() != QtCore.Qt.LeftButton:
+        if ev.button() != QtCore.Qt.MouseButton.LeftButton:
             ev.ignore()
             return
         if not self.term.is_connectable:
@@ -453,10 +456,10 @@ class TerminalGraphicsItem(GraphicsObject):
                 self.newConnection.set_target(self.mapToView(ev.pos()))
 
     def hoverEvent(self, ev):
-        if not ev.isExit() and ev.acceptDrags(QtCore.Qt.LeftButton):
-            ev.acceptClicks(QtCore.Qt.LeftButton)
+        if not ev.isExit() and ev.acceptDrags(QtCore.Qt.MouseButton.LeftButton):
+            ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton)
             # we don't use the click, but we also don't want anyone else to use it.
-            ev.acceptClicks(QtCore.Qt.RightButton)
+            ev.acceptClicks(QtCore.Qt.MouseButton.RightButton)
             self.box.setBrush(fn.mkBrush("w"))
         else:
             self.box.setBrush(self.brush)
@@ -474,7 +477,7 @@ class TerminalGraphicsItem(GraphicsObject):
 class ConnectionItem(GraphicsObject):
     def __init__(self, source, target=None):
         GraphicsObject.__init__(self)
-        self.setFlags(self.ItemIsSelectable | self.ItemIsFocusable)
+        self.setFlags(self.GraphicsItemFlag.ItemIsSelectable | self.GraphicsItemFlag.ItemIsFocusable)
         self.source = source
         self.target = target
         self.length = 0
@@ -532,6 +535,15 @@ class ConnectionItem(GraphicsObject):
         start = Point(self.source.connect_point())
         if isinstance(self.target, TerminalGraphicsItem):
             stop = Point(self.target.connect_point())
+            input_term = self.target.term if self.target.term.is_input else self.source.term
+            node_name = input_term.node.name
+            linestyle_state = input_term.node.graph._state["gui_state"][node_name]["linestyle"]
+            if input_term.name in linestyle_state:
+                shape = linestyle_state[input_term.name]
+            else:
+                shape = "cubic"
+                linestyle_state[input_term.name] = shape
+            self.style["shape"] = shape
         elif isinstance(self.target, QtCore.QPointF):
             stop = Point(self.target)
         else:
@@ -563,15 +575,30 @@ class ConnectionItem(GraphicsObject):
             ev.ignore()
             return
 
-        if ev.key() == QtCore.Qt.Key_Delete or ev.key() == QtCore.Qt.Key_Backspace:
+        if ev.key() == QtCore.Qt.Key.Key_Delete or ev.key() == QtCore.Qt.Key.Key_Backspace:
+            if self.source.term.node_type == "engine_node":
+                ev.ignore()
+                return
             self.source.disconnect(self.target)
             ev.accept()
-        elif ev.key() == QtCore.Qt.Key_Control:
+            if isinstance(self.target, TerminalGraphicsItem):
+                # Remove linestyle from gui state on removal
+                input_term = self.target.term if self.target.term.is_input else self.source.term
+                node_name = input_term.node.name
+                if input_term.name in input_term.node.graph._state["gui_state"][node_name]["linestyle"]:
+                    input_term.node.graph._state["gui_state"][node_name]["linestyle"].pop(input_term.name)
+        elif ev.key() == QtCore.Qt.Key.Key_Control:
             ev.accept()
             if self.style["shape"] == "line":
-                self.setStyle(shape="cubic")
+                shape = "cubic"
             elif self.style["shape"] == "cubic":
-                self.setStyle(shape="line")
+                shape = "line"
+            if isinstance(self.target, TerminalGraphicsItem):
+                # Update linestyle in gui state
+                input_term = self.target.term if self.target.term.is_input else self.source.term
+                node_name = input_term.node.name
+                input_term.node.graph._state["gui_state"][node_name]["linestyle"][input_term.name] = shape
+                self.setStyle(shape=shape)
         else:
             ev.ignore()
 
@@ -579,7 +606,7 @@ class ConnectionItem(GraphicsObject):
         ev.ignore()
 
     def mouseClickEvent(self, ev):
-        if ev.button() == QtCore.Qt.LeftButton:
+        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
             ev.accept()
             sel = self.isSelected()
             self.setSelected(True)
@@ -588,13 +615,13 @@ class ConnectionItem(GraphicsObject):
                 self.update()
 
     def mouseDoubleClickEvent(self, ev):
-        if int(ev.button()) == int(QtCore.Qt.LeftButton):
+        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
             ev.accept()
             if self.connection_window is not None:
                 self.connection_window.show()
 
     def hoverEvent(self, ev):
-        if (not ev.isExit()) and ev.acceptClicks(QtCore.Qt.LeftButton):
+        if (not ev.isExit()) and ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton):
             self.hovered = True
         else:
             self.hovered = False
